import pprint
import unittest

from pmlab_lite.discovery.cut import Cut
from pmlab_lite.helper.graph import Graph
from pmlab_lite.discovery.inductive_miner import InductiveMiner
from pmlab_lite.helper.io.xes import import_from_xes
from pmlab_lite.helper.viz import dot
from pmlab_lite.log import EventLog


class TesInductiveMinerMethods(unittest.TestCase):


	def test_split_seq_cut(self):
		log = [['a', 'x', 'b', 'c'],
			   ['a', 'x', 'c', 'b'],
			   ['a', 'x', 'd', 'e'],
			   ['d', 'e'],
			   ['a', 'x', 'd', 'e', 'f', 'd', 'e']]

		lg = EventLog()

		miner = InductiveMiner(lg)

		ll, lr = miner.split_log([['a', 'x'], ['b', 'c', 'd', 'e', 'f']],
								 Cut.SEQ,
								log)


		self.assertListEqual(ll, [['a', 'x'], []])
		self.assertListEqual(lr, [['b', 'c'], ['c', 'b'], ['d', 'e'],
								  ['d', 'e', 'f', 'd', 'e']])


	def test_split_max_seq_cut(self):
		log = [['a', 'x', 'b', 'c'],
			   ['a', 'x', 'c', 'b'],
			   ['a', 'd', 'e'],
			   ['a', 'x', 'd', 'e', 'f', 'd', 'e']]

		miner = InductiveMiner(EventLog())

		max_seq_split = [['a'], ['x'], [['b', 'c'], ['d', 'e', 'f']]]

		logs = miner.split_log(max_seq_split, Cut.SEQ, log)
		print(logs)

	def test_split_para_cut(self):
		log = [['a', 'b'],
			   ['b', 'a'],
			   ['a']]

		miner = InductiveMiner(EventLog())

		ll, lr = miner.split_log([['a'], ['b']], Cut.PARA, log)

		self.assertListEqual(ll, [['a']])
		self.assertListEqual(lr, [['b'], []])

	def test_split_excl_cut(self):
		log = [['a', 'b'],
			   ['d', 'e', 'f', 'd','e'],
			   ['d', 'e']]

		miner = InductiveMiner(EventLog())

		ll, lr = miner.split_log([['a', 'b'], ['d', 'e', 'f']], Cut.EXCL, log)

		self.assertListEqual(ll, [['a', 'b']])
		self.assertListEqual(lr, [['d', 'e', 'f', 'd','e'],
								  ['d', 'e']])

	def test_split_loop_cut(self):
		log = [['d', 'e', 'f', 'g', 'd', 'e'],
			   ['d', 'e']]

		miner = InductiveMiner(EventLog())

		ll, lr = miner.split_log([['d', 'e'], ['f', 'g']], Cut.LOOP, log)

		self.assertListEqual(ll, [['d', 'e']])
		self.assertListEqual(lr, [['f', 'g'], []])

	def test_prepare_log(self):

		lg = EventLog()

		lg.add_trace(1, ['a', 'b', 'g', 'h', 'i', 'j', 'k'])
		lg.add_trace(2, ['a', 'b', 'g', 'h', 'j', 'i', 'k'])
		lg.add_trace(3, ['a', 'b', 'g', 'h', 'i', 'k'])
		lg.add_trace(4, ['a', 'b', 'g', 'h', 'j', 'k'])
		lg.add_trace(5, ['a', 'b', 'c', 'd', 'e', 'h', 'i', 'k'])
		lg.add_trace(6, ['a', 'b', 'c', 'd', 'e', 'f', 'c', 'd',
						 'e', 'h', 'i', 'k'])

		miner = InductiveMiner(lg)

		result = [	['a', 'b', 'g', 'h', 'i', 'j', 'k'],
					['a', 'b', 'g', 'h', 'j', 'i', 'k'],
					['a', 'b', 'g', 'h', 'i', 'k'],
					['a', 'b', 'g', 'h', 'j', 'k'],
					['a', 'b', 'c', 'd', 'e', 'h', 'i', 'k'],
					['a', 'b', 'c', 'd', 'e', 'f', 'c', 'd', 'e', 'h', 'i', 'k']]

		result.sort()
		miner.log.sort()

		self.assertListEqual(miner.log, result)

	def test_simple_example(self):
		lg = EventLog()

		lg.add_trace(1, ['a', 'b', 'g', 'h', 'i', 'j', 'k'])
		lg.add_trace(2, ['a', 'b', 'g', 'h', 'j', 'i', 'k'])
		lg.add_trace(3, ['a', 'b', 'g', 'h', 'i', 'k'])
		lg.add_trace(4, ['a', 'b', 'g', 'h', 'j', 'k'])
		lg.add_trace(5, ['a', 'b', 'c', 'd', 'e', 'h', 'i', 'k'])
		lg.add_trace(6, ['a', 'b', 'c', 'd', 'e', 'f', 'c', 'd',
						 'e', 'h', 'i', 'k'])

		miner = InductiveMiner(lg)
		tree = miner.discover()

		tree.print_tree()
		dot.draw_process_tree(tree, format='pdf')


	def test_extended_example(self):
		lg = EventLog()

		lg.add_trace(1, ['submit application',  # a
						 'review application',  # b
						 'fast forwarding',  # g
						 'final review',  # h
						 'supervisor signature',  # i
						 'sign application',  # j
						 'close application'])  # k
		lg.add_trace(2, ['submit application',  # a
						 'review application',  # b
						 'fast forwarding',  # c
						 'final review',  # h
						 'sign application',  # j
						 'supervisor signature',  # i
						 'close application'])  # k
		lg.add_trace(3, ['submit application',  # a
						 'review application',  # b
						 'fast forwarding',  # g
						 'final review',  # h
						 'supervisor signature',  # i
						 'close application'])  # k
		lg.add_trace(4, ['submit application',  # a
						 'review application',  # b
						 'fast forwarding',  # g
						 'final review',  # h
						 'sign application',  # j
						 'close application'])  # k
		lg.add_trace(5, ['submit application',  # a
						 'review application',  # b
						 'check documents',  # c
						 'check financial status',  # d
						 'write report',  # e
						 'final review',  # h
						 'supervisor signature',  # i
						 'close application'])  # k
		lg.add_trace(6, ['submit application',  # a
						 'review application',  # b
						 'check documents',  # c
						 'check financial status',  # d
						 'write report',  # e
						 'reject report',  # f
						 'check documents',  # c
						 'check financial status',  # d
						 'write report',  # e
						 'final review',  # h
						 'supervisor signature',  # i
						 'close application'])  # k

		miner = InductiveMiner(lg)
		tree = miner.discover()

		g = Graph()
		g = g.from_log(miner.log)
		d = dot.draw_graph(g)
		dot.render_dot(d, 'dfg_process_tree')

		tree.print_tree()
		dot.draw_process_tree(tree, 'example', format='pdf')

	def test_bpi_12(self):
		log = import_from_xes('BPI_Challenge_2012.xes')

		miner = InductiveMiner(log)
		tree = miner.discover()


		tree.print_tree()
		dot.draw_process_tree(tree, 'bpi12', format='pdf')

	def test_loop_of_one(self):
		lg = EventLog()
		lg.add_trace(1, ['a', 'b', 'b', 'c'])
		lg.add_trace(2, ['a', 'c'])

		miner = InductiveMiner(lg)
		tree = miner.discover()

		print()
		tree.print_tree()
		dot.draw_process_tree(tree, 'single_loop', format='pdf')


	def test_tree_to_net(self):
		lg = EventLog()

		lg.add_trace(1, ['a', 'b', 'g', 'h', 'i', 'j', 'k'])
		lg.add_trace(2, ['a', 'b', 'g', 'h', 'j', 'i', 'k'])
		lg.add_trace(3, ['a', 'b', 'g', 'h', 'i', 'k'])
		lg.add_trace(4, ['a', 'b', 'g', 'h', 'j', 'k'])
		lg.add_trace(5, ['a', 'b', 'c', 'd', 'e', 'h', 'i', 'k'])
		lg.add_trace(6, ['a', 'b', 'c', 'd', 'e', 'f', 'c', 'd',
						 'e', 'h', 'i', 'k'])

		miner = InductiveMiner(lg)
		tree = miner.discover()

		tree.print_tree()
		# dot.draw_process_tree(tree, format='pdf')

		net = miner.tree_to_petri_net(tree)
		dot.draw_petri_net(net, 'process_tree_to_net')


from . pn import PetriNet
import numpy as np

class SynchronousProduct(PetriNet):
	""" 
	Class to represent a synchronous prodcut.
	For creation the constructor needs to get passed a PetriNet and a TraceNet.
	"""

	def __init__(self, petri_net, trace_net):
		self.places = {}
		self.transitions = {}
		self.edges = []
		self.marking = []
		self.capacity = []
		self.counter = 0  # mapping
		self.__synchronous_product(petri_net, trace_net)

	#TODO clunky -> more easy sync transitions with transitions_by_index()
	def __synchronous_product(self, petri_net, trace_net):
		#petri_net is model_net
		place_offset = len(petri_net.places.values())
		transition_offset = len(petri_net.transitions_by_index())

		#PLACES
		#copying the model net
		for p in petri_net.places.values():
			self.add_place(p)

		#copying the trace net
		for p in trace_net.places.values():
			self.add_place(p + place_offset)
			

		#TRANSITIONS
		#copying the modelnet
		model_transitions_by_index = petri_net.transitions_by_index()
		for i in range(0, len(model_transitions_by_index)):
			self.add_transition(model_transitions_by_index[i] + "_model")
		
		#copying the trace net
		trace_transitions_by_index = trace_net.transitions_by_index()
		for i in range(0, len(trace_transitions_by_index)):
			self.add_transition(trace_transitions_by_index[i] + "_log")


		#EDGES	
		#copying the model net
		for edge in petri_net.edges:
			self.add_edge(edge[0], edge[1])
			
		# copying the trace net
		for edge in trace_net.edges:
			new_edge = (0,0)
			if edge[0] > 0:
				new_edge = (edge[0]+place_offset, edge[1] - transition_offset)
			elif edge[0] < 0:
				new_edge = (edge[0] - transition_offset, edge[1]+place_offset)
			self.add_edge(new_edge[0], new_edge[1])
		
		
		#CREATE NEW SYNCHRONOUS PRODUCT TRANSITIONS AND EDGES
		#whenever trace_t has the same name as model_t we create a new sync_t with all the in/outputs from the model_ and trace_t combined
		for keyT1 in trace_net.transitions.keys():
			for keyT2 in petri_net.transitions.keys():
				if keyT1 == keyT2:
					for i in range(0, len(trace_net.transitions[keyT1])):
						keyT3 = keyT1 + "_synchronous"
						self.add_transition(keyT3)
						#copy all the in/outputs from the trace net transitions onto the new sync prod transitions
						#inputs
						for node in trace_net.get_inputs(trace_net.transitions[keyT1][i]):
							self.add_edge(node+place_offset, self.transitions[keyT3][i] )
						#outputs
						for node in trace_net.get_outputs(trace_net.transitions[keyT1][i]):
							self.add_edge(self.transitions[keyT3][i], node+place_offset)
						
						#copy all the in/outputs from the model transitions onto the new sync prod transitions
						#inputs
						for node in petri_net.get_inputs(petri_net.transitions[keyT2][0]):
							self.add_edge(node, self.transitions[keyT3][i] )
						#outpus
						for node in petri_net.get_outputs(petri_net.transitions[keyT2][0]):
							self.add_edge(self.transitions[keyT3][i], node)
		
		return self

	def get_spnets_initial_marking(self) -> list:
		#only works for nets with 2 start places so far (assuming sp-nets)
		index_start_places = self.get_index_initial_places()
		index_place_start = index_start_places[0]
		index_place_start_log = index_start_places[1]
		init_mark_vector = list (np.repeat(0, len(self.places)))
		init_mark_vector[index_place_start] = 1
		init_mark_vector[index_place_start_log] = 1
		return init_mark_vector
	
	def get_spnets_final_marking(self) -> list:
		#only works for nets with 2 end places so far (assuming sp-nets)
		index_final_places = self.get_index_final_places() 
		index_place_end = index_final_places[0]
		index_place_end_log = index_final_places[1]	
		final_mark_vector = list (np.repeat(0, len(self.places)))
		final_mark_vector[index_place_end] = 1
		final_mark_vector[index_place_end_log] = 1
		return final_mark_vector



import random
import string
from copy import copy

from pmlab_lite.discovery.cut import Cut
from pmlab_lite.helper.graph import Graph


class ProcessTree():

	def __init__(self, miner, log: list, level: int = 0,
				 parent='root', discover = True):

		self.miner = miner
		self.level = level
		self.parent = parent
		self.id = self.gen_id()
		self.log = copy(log)
		self.dfg = Graph().from_log(log)
		self.scc = self.dfg.get_scc()
		self.cc = self.dfg.get_cc()
		self.cut = None
		self.children = []

		print('---------')
		print('(tree) number of cc: %s' %len(self.cc))
		print('(tree) number of scc: %s' %len(self.scc))

		if discover:
			self.find_cuts()

	def gen_id(self) -> str:
		"""
		Returns random generated id.
		:return: String
		"""
		id = [str(self.level)]
		for i in range(4):
			id.append(random.choice(string.ascii_letters))
		return ''.join(id)

	def find_cuts(self):

		def insert_base_case(activity):
			if activity == []:
				return 'tau'
			else:
				return activity[0]

		seq_cut = self.find_seq()
		para_cut = self.find_para()
		excl_cut = self.find_excl()
		loop_cut = self.find_loop()

		print()
		print('(tree) found seq: %s, para: %s, excl: %s, loop: %s'
			  %(seq_cut, para_cut, excl_cut, loop_cut))

		if len(seq_cut) > 0:
			self.cut = Cut.SEQ
			sub_logs = self.miner.split_log(seq_cut, Cut.SEQ, self.log)
			print('(tree) seq. cut %s' %(sub_logs))


			for sub_log in sub_logs:
				if len(sub_log) > 1:
					self.children.append(ProcessTree(self.miner, sub_log,
													 self.level + 1, self))
				else:
					self.children.append(insert_base_case(sub_log[0]))

		elif len(excl_cut) > 0:
			self.cut = Cut.EXCL
			log_left, log_right = self.miner.split_log(excl_cut, Cut.EXCL,
													   self.log)

			print('(tree) excl. cut %s / %s' %(log_left, log_right))

			for sub_log in [log_left, log_right]:
				if len(sub_log) > 1:
					self.children.append(ProcessTree(self.miner, sub_log,
													 self.level + 1, self))
				else:
					self.children.append(insert_base_case(sub_log[0]))

		elif len(para_cut) > 0:
			self.cut = Cut.PARA
			log_left, log_right = self.miner.split_log(para_cut, Cut.PARA,
													   self.log)

			print('(tree) para. cut %s / %s' % (log_left, log_right))

			for sub_log in [log_left, log_right]:
				if len(sub_log) > 1:
					self.children.append(ProcessTree(self.miner, sub_log,
													 self.level + 1, self))
				else:
					self.children.append(insert_base_case(sub_log[0]))

		elif len(loop_cut) > 0:
			self.cut = Cut.LOOP
			log_left, log_right = self.miner.split_log(loop_cut, Cut.LOOP,
													self.log)

			print('(tree) loop cut %s / %s' % (log_left, log_right))


			self.children.append(ProcessTree(self.miner, log_left,
											 self.level + 1, self))

			self.children.append(ProcessTree(self.miner, log_right,
											 self.level + 1, self))

		else:
			print('(tree) base case: %s' %self.log)
			if len(self.log) > 1:
				if all([len(set(sub_log)) == 1 for sub_log in self.log]):
					self.cut = Cut.EXCL
					for ele in self.log:
						self.children.append(insert_base_case(ele))
				else:
					self.cut = Cut.EXCL
					for sub_log in self.log:

						if len(sub_log) < 2:
							self.children.append(insert_base_case(sub_log))
						else:
							tree = ProcessTree(self.miner, sub_log,
											   self.level + 1, self.id, False)
							tree.cut = Cut.LOOP
							tree.children.append(insert_base_case(sub_log))
							self.children.append(tree)
			else:
				self.children.append(insert_base_case(sub_log))

	def find_seq(self):
		print('(tree) search for SEQ')

		if(len(self.scc) > 1 and len(set(self.dfg.start_nodes))) == 1:
			if(len(self.cc) == 1):

				mapping = {}
				for comp in self.scc:
					mapping[len(mapping)] = comp

				def get_mapping(value):
					for k, v in mapping.items():
						if value[0] in v:
							return k
					return None

				sorted_scc = {}
				scc = self.scc  # TODO don't use mapping
				# print(scc)
				g = self.dfg  # TODO don't use mapping
				# pprint.pprint(g.vertexes)

				for comp in scc:
					sorted_scc[get_mapping(comp)] = [[], [], []]
					for comp2 in scc:
						if comp is not comp2:
							if g.is_reachable(comp[0], comp2[0]) and \
									not g.is_reachable(comp2[0], comp[0]):
								# comp before comp2
								sorted_scc[get_mapping(comp)][0].append(comp2)
							elif not g.is_reachable(comp[0], comp2[0]) and \
									g.is_reachable(comp2[0], comp[0]):
								# comp after comp2
								sorted_scc[get_mapping(comp)][1].append(comp2)
							else:
								# either parallel or exclusive
								sorted_scc[get_mapping(comp)][2].append(comp2)


				# find very first activity
				first = []
				for key, relations in sorted_scc.items():
					if len(relations[1]) == 0:
						first.append(key)

				# get all succeeding events in no order
				first = mapping[first[0]]
				next_candidates = []
				for key, relations in sorted_scc.items():
					if first in relations[1]:  # [before][after][para/excl]
						next_candidates.append(mapping[key])


				# sort succeeding events
				pre = [first]
				while (len(next_candidates) > 0):
					add = True
					for candidate in next_candidates:
						for key, relations in sorted_scc.items():
							if candidate in relations[0]:
								if mapping[key] not in pre:
									add = False
						if add:
							pre.append(candidate)
							next_candidates.remove(candidate)

				result = []
				for idx, comp in enumerate(pre):
					print(comp)
					if len(sorted_scc[get_mapping(comp)][2]) > 0:
						if idx > 0 and \
								len(sorted_scc[get_mapping(pre[idx - 1])][2]) > 0:
							result[-1].append(comp)
						else:
							result.append([comp])
					else:
						result.append(comp)
				return result
			else:
				print('(tree) No sequence -> multiple CC OR multiple start '
					  'points')
				return []
		else:
			print('(tree) No sequence -> single SCC')
			return []

	def find_para(self):
		print('(tree) search for PARA')
		g_i = self.dfg.invert()
		scc_i = g_i.get_scc()

		candidates = []
		for o in scc_i:
			for t in scc_i:
				if o[0] != t[0]:
					if self.dfg.is_reachable(o[0], t[0]) and \
							self.dfg.is_reachable(t[0], o[0]):
						# TODO: maybe rework this
						if o[0] in g_i.start_nodes and t[0] in g_i.start_nodes:
							if o not in candidates:
								candidates.append(o)
							if t not in candidates:
								candidates.append(t)

		return candidates

	def find_excl(self):
		print('(tree) search for EXCL')
		for o in self.scc:
			for t in self.scc:
				if o[0] != t[0]:
					if not self.dfg.is_reachable(o[0], t[0]) and \
							not self.dfg.is_reachable(t[0], o[0]):
						return [o, t]
		return []

	def find_loop(self):
		print('(tree) search for LOOP')


		g_i = self.dfg.invert()

		g_reduced = g_i.remove_start()
		g_reduced = g_reduced.remove_end()
		branches = g_reduced.get_scc()
		if len(branches) == 0:
			return []
		elif len(branches) == 1:
			backward = branches[0]
			forward = [self.dfg.start_nodes[0], self.dfg.end_nodes[0]]
			return [forward, backward]
		else:
			backward = []
			forward = []
			for b in branches:
				if b[0] in self.dfg.vertexes[self.dfg.start_nodes[0]]:
					if isinstance(b, list):
						forward += b
					else:
						forward.append(b)
				elif b[0] in self.dfg.vertexes[self.dfg.end_nodes[0]]:
					backward = b
				else:
					backward += b
			forward = [self.dfg.start_nodes[0]] + forward \
					  + [self.dfg.end_nodes[0]]

			return [forward, backward]

	def print_tree(self):
		"""
		Print tree structure to console.
		"""

		def print_cut(cut: int) -> str:
			if cut == Cut.SEQ:
				return 'SEQ'
			elif cut == Cut.PARA:
				return 'PAR'
			elif cut == Cut.EXCL:
				return 'EXL'
			elif cut == Cut.LOOP:
				return 'LOOP'

		if isinstance(self.parent, ProcessTree):
			parent = self.parent.id
		else:
			parent = 'root'
		print('%s (%s/%s) node: %s children: %s' %
			  (self.level, parent, self.id,
			   print_cut(self.cut), self.children))

		for child in self.children:
			if isinstance(child, ProcessTree):
				child.print_tree()


import unittest

from pmlab_lite.helper.graph import Graph
import pmlab_lite.log as eLog

class TestGraphMethods(unittest.TestCase):

	def test_from_log(self):

		log = eLog.EventLog()
		log.add_trace(1, ['A', 'B', 'C', 'E'])
		log.add_trace(2, ['A', 'B', 'D', 'E'])

		g = Graph()

		g.from_event_log(log)

		solution = {'A': ['B'],
					'B': ['C', 'D'],
					'C': ['E'],
					'D': ['E'],
					'E': []}

		for k, v in g.graph.items():
			self.assertListEqual(sorted(v), sorted(solution[k]))

	def test_find_scc(self):

		g = Graph()
		g.add_edge('a', 'b')
		g.add_edge('b', 'c')
		g.add_edge('c', 'd')
		g.add_edge('d', 'b')
		g.add_edge('c', 'e')
		g.add_edge('e', 'f')
		g.add_edge('f', 'g')
		g.add_edge('g', 'h')
		g.add_edge('h', 'f')
		g.add_edge('g', 'i')
		g.add_edge('i', 'j')
		g.add_edge('j', 'i')
		g.add_edge('j', 'k')
		g.add_edge('k', 'l')


		# -----

		solution = [['a'], ['b','c','d'], ['e'], ['f','g','h'], ['i','j'],
					['k'], ['l']]


		c = g.get_scc()

		
		self.assertEqual(len(c), len(solution))

		for comp1 in c:
			exisits = False
			for comp2 in solution:
				if sorted(comp1) == sorted(comp2):
					exisits = True
			self.assertTrue(exisits)


import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name='pmlab_lite',
    version='0.4.2.2',
    author=['Josep Carmona',
			'Matthias Weidlich',
			'Andrea Burattin',
			'jmunozgama',
			'Simon Remy',
			'Glenn Dittmann'],
    author_email='',
    description='Process Mining scripting environment',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/pmlab/pmlab-lite',
    packages=setuptools.find_packages(),
    license='GPL-3.0',
	install_requires=[
    		'numpy',
    		'graphviz',
    		'tqdm',
    	], #maybe pythone-dateutil
	classifiers=[
			'Development Status :: 2 - Pre-Alpha',
			'Programming Language :: Python :: 3',
		  ],
)



from . pn import PetriNet



class TraceNet(PetriNet):
	""" 
	Class to represent a synchronous prodcut.
	For creation the constructor needs to get passed a trace.
	"""

	def __init__(self, trace: list):
		self.places = {}
		self.transitions = {}
		self.edges = []
		self.marking = []
		self.capacity = []
		self.counter = 0  # mapping
		self.make_trace_net(trace)

	def make_trace_net(self, trace: list):
		"""
		Takes a trace, as a list of strings(=events) and makes a traces net from it, 
		i.e. a sequential connection of the events as places with transitions in between.
		"""
		#assume empty PetriNet
		num_places = len(trace)+1

		for i in range(1, num_places+1):
			self.add_place(i)

		for t in trace:
			self.add_transition(t)
 
		for i in range(1, num_places):
			self.add_edge(i, -i)
			self.add_edge(-i, i+1)

		return self


#class to store te specifications if the found alignment(s)
class Alignment:
	def __init__(self):
		self.alignment_move = []
		self.move_in_model = []
		self.move_in_log = []
		self.fitness = []
		self.closed_list_end = []



from . abstract_pn import AbstractPetriNet
from random import randint
import pprint
import numpy as np


class PetriNet(AbstractPetriNet):
	""" Class to represent a Petri Net."""

	def add_place(self, place_name: int, capacity: int=1):
		"""
		Add a place to the net and set its token capacity. The name
		has to be numeric. Further, the keys are indices of the corresponding
		marking vector and capacity vector.

		Args:
			name: integer place name
			capacity: integer number of token per place

		Raises:
			ValueError: place identifier has to be unique
			TypeError: place identifier has to be numeric
		"""
		if isinstance(place_name, int) and place_name > 0:
			if not self.place_exists(place_name):
				idx = len(self.places)
				self.places[idx] = place_name
				self.marking.append(0)
				self.capacity.append(capacity)
			else:
				raise ValueError('place identifier has to be unique')
		else:
			raise TypeError('place identifier has to be numeric and > 0')

		return self

	def remove_place(self, place_name: int): 
		"""
		Remove a place from the net.

		Args:
			name: integer place name
		"""
		index = 0   #!!!if name not in places -> index stays 0 and first place gets removed..

		for idx, place in self.places.items():
			if place == place_name:
				index = idx
				break

		# shift places with larger index to the left
		if index + 1 == len(self.places):  # last index
			self.places.pop(index)
		else:
			for idx in range(index, len(self.places) - 1):
				self.places[idx] = self.places[idx + 1]

			self.places.pop(len(self.places) - 1)

		del self.marking[index]
		del self.capacity[index]

		return self

	def num_places(self) -> int:
		return len( self.places.keys() )

	def get_mapping(self) -> dict:
		return self.transitions

	def get_marking(self) -> list:
		return self.marking

	def add_transition(self, name: str, transition_id: int=None):
		"""
		Add a transition to the net. The name has to be a string.
		Adding the same name multiple times will append to the list which the name(key) points to.

		Args:
			name: string name of transition, key 
			id: negative integer, value of transition at the key(name)
		"""

		self.counter -= 1

		if len(name) == 0:
			name = 'tau'

		if transition_id is None:
			transition_id = self.counter
		else:
			if transition_id >= 0:
				raise ValueError('transition identifier has to be < 0')

		if name in self.transitions.keys():
			self.transitions[name].append(transition_id)
		else:
			self.transitions[name] = [transition_id]

		return self

	def remove_transition(self, id: int):
		"""
		Remove the given transition from the net.

		Args:
			id: value of the transition at key=name
		"""

		for key, values in self.transitions.items():
			if id in values:
				values.remove(id)
				if len(values) == 0:
					self.transitions.pop(key, None)
				break

		return self
		
	def num_transitions(self) -> int:
		return sum( [len(v) for v in self.transitions.values()] )

	def add_edge(self, source, target, two_way=False):
		#could also get one touple as an argument...
		"""
		Add a new edge between two elements in the net. If two way argument
		is True, one edge per direction will be added.

		Args:
			source: id of transition/name of place
			target: id of transition/name of place
			two_way: add edge from target to source

		Raises:
			ValueError: source/target does not exists
		"""
		if source > 0 and target > 0 or source < 0 and target < 0:
			raise ValueError('edges can only be added between places and '
							 'transition and vice versa')

		if source > 0:
			# source is place
			if not self.place_exists(source):
				raise ValueError('place does not exist')
		else:
			# source is transition
			if not self.transition_exists(source):
				raise ValueError('transition does not exist')

		if target > 0:
			# target is place
			if not self.place_exists(target):
				raise ValueError('place does not exist')
		else:
			# target is transition
			if not self.transition_exists(target):
				raise ValueError('transition does not exist')

		if not two_way:
			self.edges.append((source, target))
		else:
			self.edges.append((source, target))
			self.edges.append((target, source))

		return self

	def remove_edge(self, source, target):
		"""Remove edge.

		Args:
			source: id of transition/name of place
			target: id of transition/name of place
		"""
		for idx, edge in enumerate(self.edges):
			if edge[0] == source and edge[1] == target:
				del self.edges[idx]
				break

		return self

	def remove_all_edges_of(self, name):
		"""
		Remove all edge where the given element is either the source or
		the target.

		Args:
			name: id of transition/name of place
		"""
		# where element is source
		for e in list(filter(lambda x: x[0] == name, self.edges)):
			del self.edges[self.edges.index(e)]

		# where element is target
		for e in list(filter(lambda x: x[1] == name, self.edges)):
			del self.edges[self.edges.index(e)]

		return self

	def index_of_place(self, place_name: int) -> int:
		for idx, p in self.places.items():
			if p == place_name:
				return idx

	def transitions_by_index(self) -> dict:
		"""
		Returns the reverse of data structure of the transitions, 
		i.e. a dict where the id of the transitions are the keys and the
		values are the transtions names (strings)
		"""
		transitions_by_index = dict()

		for key, value in self.transitions.items():
				for val in value:
					transitions_by_index[-(val+1)] = key

		return transitions_by_index

	def is_enabled(self, transition_id: int) -> bool:
		"""
		Check whether a transition is able to fire or not.

		Args:
			transition: id of transition

		Retruns:
			True if transition is enabled, False otherwise.
			Special case: returning true, when a transition has no input places
		"""

		# all palces which are predecessor of the given transition
		inputs = self.get_inputs(transition_id)
		# do any inputs exist?
		if len(inputs) > 0:
			# check if each place contains at least one token aka. has a
			# marking
			for i in inputs:
				idx = self.index_of_place(i)
				# input place has no token
				if self.marking[idx] == 0:
					return False
			# transition is able to fire
			return True
		else:
			# no input places
			return True

	def get_inputs(self, node: int) -> list:
		"""
		args: node; name of place or id of transition

		returns: list of numbers, positives when node is a transition and negatives when node is a place
		"""
		inputs = []
		for edge in self.edges:
			if edge[1] == node:
				inputs.append(edge[0])
		return inputs

	def get_outputs(self, node: int) -> list:
		"""
		args: node; name of place or id of transition

		returns: list of numbers, positives when node is a transition and negatives when node is a place
		"""
		outputs = []
		for edge in self.edges:
			if edge[0] == node:
				outputs.append(edge[1])
		return outputs

	def add_marking(self, place_name: int, num_token: int=1):
		"""
		Add a new marking to the net.

		Args:
			place: name of place
			num_token: number of token to add

		Raises:
			ValueError: place does not exists
		"""
		index = None
		if self.place_exists(place_name):
			for idx, p in self.places.items():
				if p == place_name:
					index = idx
					break

			self.marking[index] = num_token
		else:
			raise ValueError('place does not exist.')

		return self

	def null_marking(self):
		"""
		Resets the net to the null marking, i.e. all places contain no token
		"""
		for i in range(0, len(self.marking)):
			self.marking[i] = 0

	def replay(self, max_length: int) -> list:
		"""Randomly replay the net starting from its current marking.

		Returns:
			Sequence of transitions which were fired.
		"""
		seq = []
		enabled_transitions = self.all_enabled_transitions()
		while (len(enabled_transitions) > 0 and len(seq) < max_length):
			t = enabled_transitions[randint(0, len(enabled_transitions) - 1)]
			seq.append(t)
			self.fire_transition(t)
			enabled_transitions = self.all_enabled_transitions()

		if len(seq) == max_length:
			seq.append('BREAK')

		return seq

	def transition_exists(self, transition_id: int) -> bool:
		"""
		Check whether the transition exists in the net or not.

		Args:
			name: id of transition

		Returns:
			True if transition exists in petri net, False otherwise.
		"""

		# flatten list
		transition_mapping = [item for sublist in self.transitions.values() for
							  item
							  in sublist]
		return transition_id in transition_mapping

	def place_exists(self, place_name: int) -> bool:
		"""
		Check whether the place exists in the net or not.

		Args:
			name: name of place

		Returns:
			True if place exists in petri net, False otherwise.
		"""

		return place_name in list(self.places.values())

	def all_enabled_transitions(self) -> list:
		"""
		Find all transitions in the net which are enabled.

		Returns:
			List of all enabled transitions.
		"""
		transitions = [item for sublist in self.transitions.values() for item
					   in sublist]

		return list(filter(lambda x: self.is_enabled(x), transitions))

	def fire_transition(self, transition_id: int):
		"""
		Fire transition.

		Args:
			name: id of transition
		"""
		if self.is_enabled(transition_id):
			inputs = self.get_inputs(transition_id)

			outputs = self.get_outputs(transition_id)

			# update ingoing token
			for i in inputs:
				idx = self.index_of_place(i)
				self.marking[idx] -= 1

			# update outgoing token
			for o in outputs:
				idx = self.index_of_place(o)
				self.marking[idx] += 1
		else:
			print("Transition is not enabled!")

	def __repr__(self):
		"""
		Change class representation.

		:return: string
		"""
		desc = "Transitions: %s \n" \
			   "Places: %s \n" \
			   "Capacities: %s \n" \
			   "Marking: %s \n" \
			   "Edges: %s" % (self.transitions, self.places, self.capacity,
							  self.marking, self.edges)

		return desc

	def get_index_initial_places(self) -> list:
		"""
		Returns a list of the indices of the initial places, i.e. the keys to the values(names) of the place names for all places,
		who do not have any input transitions.
		"""
		index_places_start = []
		for key in self.places.keys():
			if len(self.get_inputs(self.places[key])) == 0:
				index_places_start.append(key)
		return index_places_start

	def get_index_final_places(self) -> list:
		"""
		Returns a list of the indices of the final places, i.e. the keys to the values(names) of the place names for all places,
		who do not have any output transitions.
		"""
		index_places_end = []
		for key in self.places.keys():
			if len(self.get_outputs(self.places[key])) == 0:
				index_places_end.append(key)
		return index_places_end
	
	def incidence_matrix(self):
		# Creating an empty matrix							  	
		incidence_matrix = np.zeros( ( self.num_places(), self.num_transitions() ), dtype=int)
        
		transitions_by_index = self.transitions_by_index()

		for t in transitions_by_index:
			for key in self.places.keys():
				t_val = -(t+1)					#reverse the index, to be the transtions value again
				col_index = t
				row_index = key
				p = self.places[key]
				#edge goes from P to T and vice versaa
				if ( (t_val, p) in self. edges) and ( (p, t_val) in self. edges): 
					incidence_matrix[row_index][col_index] = 0
				#edge goes from T to P
				elif (t_val, p) in self.edges:
					incidence_matrix[row_index][col_index] = 1
				#edge goes from P to T
				elif (p, t_val) in self.edges:
					incidence_matrix[row_index][col_index] = -1

		return incidence_matrix